/*
	Intermediate model format for game meshes. GameMesh
	www.janvanderweg.com
	jan@janvanderweg.com
	
	- Bone hierarchy subsection
	
	references:
	Lengyel, Eric. Computing Tangent Space Basis Vectors for an Arbitrary Mesh. Terathon Software 3D Graphics Library, 2001. http://www.terathon.com/code/tangent.html
*/ 

fn gmCollectHierarchy GS=
(
	numnodes = GS.NumNodes
	format "--- Collecting hierarchy for % nodes---\n" numnodes
	
	GS.Hierarchy = gmHierarchy()
	GS.Hierarchy.isHierarchy = false 
	local Allbones=#()
	local rootbones=#()
	
	max modify mode	-- Must set max to modify mode so the script can acess the skin modifier
	for i = 1 to numnodes do
	(
		local skinmod
		node = GS.gmNodes[i].MaxNode
		mesh = GS.gmNodes[i].Mesh
		_gmNode = GS.gmNodes[i]
		format "\tNode: %\n" node
		--- Find if node is using skinmod
		mods = node.modifiers
		for i = 1 to mods.count do
		(
			if classof mods[i] == Skin do
			(
				if skinmod == undefined do -- get first found skinmodifier and hope its on top of the stack
				(
					skinmod = mods[i]
					GS.Hierarchy.isHierarchy = true
				)
			)
		)
		if (skinmod == undefined) then
		(
			/* No skinmod found but we still want to assign the verts to a bone.
				- if it is linked to a bone assign verts to that bone.
				- if it isn't linked to a bone assign to root
			*/
			parent = node.parent
			if parent != undefined and (classof parent) == BoneGeometry then
			(
				GS.Hierarchy.isHierarchy = true
				for v = 1 to mesh.numVerts do
				(
					--- Gather vertex weight values and bones
					weightvalues=#()
					weightbones=#()
					
					append weightvalues 1.0
					append weightbones parent.name
					
					-- padd weights with null values untill size is equal to max vertex weight count
					for i =1 to (GS.setting_maxweightspervertex - 1) do
					(
						append weightvalues 0.0
						append weightbones ""	-- does an empty string work with max?
					)
					
					append _gmNode.VertexWeightValues weightvalues
					append _gmNode.VertexWeightBones weightbones
				)
				_gmNode.LinkType = 1
			)
			else
			(
				/*
				for v = 1 to mesh.numVerts do
				(
					--- Gather vertex weight values and bones
					weightvalues=#()
					weightbones=#()
					
					append weightvalues 1.0
					append weightbones "gmrootbone"
					
					-- padd weights with null values untill size is equal to max vertex weight count
					for i =1 to (GS.setting_maxweightspervertex - 1) do
					(
						append weightvalues 0.0
						append weightbones ""	-- does an empty string work with max?
					)
					
					append _gmNode.VertexWeightValues weightvalues
					append _gmNode.VertexWeightBones weightbones
				)*/
				_gmNode.LinkType = 2
			)
		)
		--if undefined != skinmod do
		else
		(
			/* The node has a skin modifier so now its time to process it all. The process:
				- Find all the bones used in the skin modifier
				- Work up all bone parents untill a root is found. Find all rootbones this way
				- Find all child bones of these parent bones. This will find bones not added to the skinmodifier
			*/
			--max modify mode	-- Must set max to modify mode so the script can acess the skin modifier
			_gmNode.LinkType = 0
			modPanel.setCurrentObject skinmod node:node
			
			local sm_bonecount = skinOps.GetNumberbones skinmod
			
			for i = 1 to sm_bonecount do
			(
				local name = skinOps.GetBoneName skinmod i 0
				if GS.setting_verbosehierarchy == true do format "Bone: %\n" name
				local bone = getNodeByName name exact:true
				append Allbones bone
			)
			for v = 1 to mesh.numVerts do
			(
				--- Gather vertex weight values and bones
				weightvalues=#()
				weightbones=#()
				weightcount = (skinOps.GetVertexWeightCount skinmod v)
				
				--- Check if vertex weight count doesn't exceed max vert weights. Default is 4
				if weightcount > GS.setting_maxweightspervertex do
				(
					GS.isError = true
					GS.whatError = "Hierarchy: Weightcount on vertex: " + v + "\tNode: " + node.name " exceeds max vertex weightcount: " + GS.setting_maxweightspervertex
					return false
				)
				
				for w = 1 to weightcount do
				(
					weight = skinOps.GetVertexWeight skinmod v w
					boneid = skinOps.GetVertexWeightBoneID skinmod v w -- skinmod internal system bone id
					bone = skinOps.GetBoneName skinmod boneid 0
					
					append weightvalues weight
					append weightbones bone
				)
				-- padd weights with null values untill size is equal to max vertex weight count
				for i = weightcount to (GS.setting_maxweightspervertex - 1) do
				(
					append weightvalues 0.0
					--append weightbones ""	-- does an empty string work with max?
					append weightbones undefined
				)
				append _gmNode.VertexWeightValues weightvalues
				append _gmNode.VertexWeightBones weightbones
			)
		)
	)
	if GS.Hierarchy.isHierarchy == true do
	(
		for i = 1 to numnodes do
		(
			if GS.gmNodes[i].LinkType == 2 do
			(
				GS.isError = true
				GS.whatError = "GameMesh is animated while node: '" + GS.gmNodes[i].maxnode.name + "' has no skinmod or is not parented to a bone."
				return false
			)
		)
	)
	max utility mode
	if GS.setting_verbosehierarchy == true do format "Final collected bone list:\n"
	for i = 1 to Allbones.count do
	(	
		if GS.setting_verbosehierarchy == true do format "\t%\n" Allbones[i].name
		local rootbone = FindRootBone Allbones[i]
		if GS.setting_verbosehierarchy == true do format "\t\t%\n" rootbone
		appendifunique rootbones rootbone
	)
	if GS.setting_verbosehierarchy == true do
	(
		format "Final rootbones: \n"
		for i = 1 to rootbones.count do
		(
			format "\t%\n" rootbones[i].name
		)
	)
	/* Create a list of bone nodes where parents come before children */
	local pA=#() -- parent array to hold parent nodes
	pA = rootbones
	local nA=#() -- final node array holding sorted nodes where parents always come before children
	local tA=#() -- temp array to hold the list of children nodes from the nodes in pA

	while pA.count > 0 do
	(
		for i = 1 to pA.count do 													--Go through each node in pA and:
		(
			--format "Classof pA[i]: %\n" (classof pA[i])
			if (classof pA[i] == BoneGeometry) do
			(
				append nA pA[i]															-- Append nodes in pA into nA
				-- append children of each node in pA into tA
				local children = pA[i].children										-- get children of node pA[i]
				for j = 1 to children.count do append tA children[j]			-- put all children in tA
			)
		)
		pA=#()																			-- clear pA
		for k = 1 to tA.count do append pA tA[k]							-- copy tA into pA
		tA=#()																			-- Clear tA for next cycle
	)
	--GS.Hierarchy.bones = #("gmrootbone")
	join GS.Hierarchy.bones na
	--GS.Hierarchy.bones = na
	
	if Gs.Hierarchy.isHierarchy == true then 
	(
		GS.isAnimated = true
		
		format "GameMesh is animated\n"
		format "Final sorted bonelist: \n"
		for i = 1 to GS.Hierarchy.bones.count do
		(
			bone = GS.Hierarchy.bones[i]
			parent = bone.parent
			
			GS.Hierarchy.count = GS.Hierarchy.bones.count
			append GS.Hierarchy.bonename bone.name
			
			if parent == undefined then append GS.Hierarchy.parentindex -1
			else
			(
				parentindex = (finditem GS.Hierarchy.bones parent)
				parentindex -= 1
				append GS.Hierarchy.parentindex parentindex
			)
			
			format "\t%\t%\t\"%\"\n" GS.Hierarchy.bones[i].name GS.Hierarchy.parentindex[i] GS.Hierarchy.bonename[i]
		)
		
		
		/*format "GameMesh is animated\n"
		format "Final sorted bonelist: \n"
		for i = 1 to GS.Hierarchy.bones.count do
		(
			format "\t%\n" GS.Hierarchy.bones[i].name
		)*/
		format "Bones:\n\n%\n\n" GS.Hierarchy
	)
	else format "GameMesh is NOT animated\n"
)