/*
	Intermediate model format for game meshes. GameMesh
	www.janvanderweg.com
	jan@janvanderweg.com
	
	- The meat of the script, the mesh processor
	
	references:
	Lengyel, Eric. Computing Tangent Space Basis Vectors for an Arbitrary Mesh. Terathon Software 3D Graphics Library, 2001. http://www.terathon.com/code/tangent.html
*/ 

fn gmCollectMeshes GS=
(
	numnodes = GS.NumNodes
	format "--- Processing % node mesh(es) ---\n" numnodes
	/* The process explained! (I hope!)
	Per node do:
		- Calculate the vertex normals for each vertex on each face
		- Iterate through all faces
		- for each vert check if it exists in the index lists.
				same vert, tvert and the vert is using an y of the given smoothing groups
		- if found just point the face vert to existing index
		- if not found add a new vert to the index lists and point to that
	*/
	
	for i = 1 to numnodes do
	(
		start = timeStamp()
		
		node = GS.gmNodes[i].MaxNode
		_gmNode = GS.gmNodes[i]
		numfaces = GS.gmNodes[i].Mesh.numFaces
		mesh = GS.gmNodes[i].Mesh
		
		format "Working node: %\n" node.name
		format "\tBuilding vertex normals... "
		/* build empty list of vertex normals for each maxvert */
		/* find bounding radius for node */
		numverts = GS.gmNodes[i].Mesh.numVerts
		for vert = 1 to numverts do
		( 
			vn = gmVertexNormals()
			append _gmNode.Normals vn

			--BOUNDING RADIUS
			radiusvert = getvert mesh vert
			local x = radiusvert[1]
			local y = radiusvert[2]
			local z = radiusvert[3]
			
			-- calculate vector length
			local length = sqrt((x^2)+(y^2)+(z^2))
			if length > _gmNode.BRadius do _gmNode.BRadius = length
		)
		--- TODO: Merge this in the vertex list generation. Keeping it in its own loop for testing/readable code purposes
		for tri = 1 to numfaces do
		(
			-- grab max index --
			vert_indices = getFace mesh tri
			uv_indices = getTVFace mesh tri
			
			for index = 1 to 3 do
			(
				nextvert = index + 1
				if nextvert > 3 do nextvert = 1
				previousvert = index - 1
				if previousvert < 1 do previousvert = 3
					
				vN = getVert mesh vert_indices[nextvert]
				vP = getVert mesh vert_indices[previousvert]
				vC = getVert mesh vert_indices[index]
				
				uvN = getTVert mesh uv_indices[nextvert]
				uvP = getTVert mesh uv_indices[previousvert]
				uvC = getTVert mesh uv_indices[index]
				
				E1 = vN - vC
				E2 = vP - vC
				
				UV1 = uvN - uvC
				UV2 = uvP - uvC 
				
				normal = getfacenormal mesh tri
				tangent = point3 ((UV2[2] * E1[1]) - (UV1[2] * E2[1])) \
										((UV2[2] * E1[2]) - (UV1[2] * E2[2])) \
										((UV2[2] * E1[3]) - (UV1[2] * E2[3]))
				bitangent = point3 ((UV1[1] * E2[1]) - (UV2[1] * E1[1])) \
											((UV1[1] * E2[2]) - (UV2[1] * E1[2])) \
											((UV1[1] * E2[3]) - (UV2[1] * E1[3]))
				/*--- Gram schmidt orthogonalize ---*/ /**(does it work?)**/
				GStangent = normalize (tangent - (normal * (dot normal tangent)))
					
				/* Add generated normals to normals struct */
				local vi = vert_indices[index]
				append _gmNode.Normals[vi].sgroup ( getFaceSmoothGroupB mesh tri)
				append _gmNode.Normals[vi].normal (normalize normal)
				append _gmNode.Normals[vi].tangent (normalize tangent)
				append _gmNode.Normals[vi].bitangent (normalize bitangent)
				append _gmNode.Normals[vi].faces tri
					
				/*format "normal index: %\n" vi
				format "\tgstangent: %\n" GStangent
				format "\t%\n\t%\n\t%\n" normal tangent bitangent*/
			)
		)
		pgstring = "Generating index lists: " + node.name
		progressstart pgstring
		progressupdate 0.0
		
		format "Generating index lists..."
		for tri = 1 to numfaces do
		(
			if (mod tri 50) == 0 do
				progressupdate (100.0*tri/numfaces)

				-- get verts and tverts used for each face index. get smooth groups used by face
			vert = getFace mesh tri
			tvert = getTVFace mesh tri
			sgused = getFaceSmoothGroupB mesh tri
			
			--format "Tri %: %\t%\t%\n" tri vert tvert sgused
			
			newfaceindices = [0,0,0]
			
			/* Optimization notes:
				- FUCK can't use bsearch since you can't sort smoothing groups in any sane way that lets bsearch actually work
				- Bah, couldn't use bsearch anyway since if works on a pre-qsorted list. Since the list is constantly growing the script would need to qsort it. Qsort is slow
				- Using finditem for the fuuuuuuuuuuuuuuuuuuuck
				- Adding a check to end the loop prematurely when it finds a match makes the whole loop slower.
					- using a for while makes it slower
					- at the start of a loop run check if found, if found stop. Also slower... :(
				- Starting search from the rear also slows it down... what? I thought this would speed it up quite a bit. Needs more research
				- Using 'and' operators when checking if the vert already exists is slower than check if vert exists, then tvert, then finally comparing sgroups.
					- Using nested if statements are noticibly faster.
				- Replaced smoothgroupsharing checks and merge functions with built in bitarray operators. Great speed up yay!
			
				Things to still check on:
				- start looking from the back of the list instead of front.
				- still use bsearch but make some fancy container so I don't have to compare smoothing groups
					- would have to find a way to insert new containers at the correct location
			*/
			--format "face: %, %, %\n" vert[1] vert[2] vert[3]
			for index = 1 to 3 do -- iterate through all verts of the face
			(
				/* look through all previous added vert/tverts/sg if the to be added vert already exists.
					if it does just point to the old index, if its new create new index
				*/
				local found = false
				
				for v = 1 to _gmNode.VertIndex.count do
				(
					/* Nested if statements are faster yay! finally an optimization that works */
					if vert[index] == _gmNode.VertIndex[v] do
					(
						if tvert[index] ==  _gmNode.TVertIndex[v] do
						(
							if ( sgused * _gmNode.SmoothGroups[v]).numberset > 0 do
							(
							/* 'upgrade' the used smoothing groups for found index since a face can share the same vert
								while also having more smoothing groups turned on. */
							--format "found: %,\t%\n" sgused _gmNode.SmoothGroups[v]
							--format "\tv: % tv: % sgused: % reqsg: %\n" vert[index] tvert[index] sgused _gmNode.SmoothGroups[v]
							_gmNode.SmoothGroups[v] = (sgused + _gmNode.SmoothGroups[v])
							found = true
							newfaceindices[index] = v
							)
						) 
					)
				)
				if found == false do
				(
					--- An existing corresponding vert was not found. Add a new entry
					newfaceindices[index] = _gmNode.VertIndex.count + 1 -- do + 1 since we're adding the new vert to the list
					append _gmNode.VertIndex vert[index]
					append _gmNode.TVertIndex tvert[index]
					append _gmNode.SmoothGroups sgused
					--format "\tadded: %\t%\t%\n" vert[index] tvert[index] sgused
				)
			)
			append _gmNode.Tris newfaceindices
			matid = (getFaceMatID mesh tri)
			_gmNode.MatIdsUsed[matid] = true
			append _gmNode.MatIds matid
		)
		format "Done in: "
		end = timeStamp()
		format "%\n" ((end - start) / 1000.0)
		
		progressEnd()
		/*for v = 1 to _gmNode.VertIndex.count do 
			format "%: %\t%\t%\n" v _gmNode.VertIndex[v] _gmNode.SmoothGroups[v] _gmNode.TVertIndex[v]
		for f = 1 to _gmNode.Tris.count do
			format "%: %\t%\n" f _gmNode.Tris[f] _gmNode.MatIds[f]*/
	
	)
)
