/*
	Intermediate model format for game meshes. GameMesh
	www.janvanderweg.com
	jan@janvanderweg.com
	
	- some base struct defs
	
	references:
	Lengyel, Eric. Computing Tangent Space Basis Vectors for an Arbitrary Mesh. Terathon Software 3D Graphics Library, 2001. http://www.terathon.com/code/tangent.html
*/ 

/* the GameMesh node container */
struct gmNode
(
	MaxNode,
	Mesh,
	Material,
	
	/* index lists */
	VertIndex=#(),
	TVertIndex=#(),
	/* A vert can be shared with multiple faces. Each face says this vert has different normals (smoothgroup none)
		store all these normals in normals.
		some normals are only used with certain smoothing groups so for each normal per vert store which smoothing group it 
		is associated with		
	*/
	Normals=#(),					-- a list of vertexnormals found on the maxvert for this vert, size = numver of maxverts
	SmoothGroups=#(),			-- which sgroups does this vert use? size = size of indexlist
	/* */
	
	VertexWeightValues=#(),
	VertexWeightBones=#(),
	Tris=#(),
	MatIds=#(),
	MatIdsUsed=#{},
	
	BMin=[0,0,0],
	BMax=[0,0,0],
	BRadius=0,
	
	TimeToBuild=0,
	LinkType=0,		/*	0 = skinmod
								1 = linked to a bone
								2 = using gmrootbone */
	
	fn DebugOut=
	(
		format "--- GameNode: ---\n"
		format "\tMaxNode: %\n" MaxNode
		format "\tMaterial: \n"
		Material.DebugOut()
		format "\tVert Indices: %\n" VertIndex.count
		format "\tTri count: %\n" Tris.count
		format "\tMaterial ids used: %\n" MatIdsUsed
		format "\tLinkType: "
		if LinkType == 0 do format "Skinmod\n"
		if LinkType == 1 do format "Parent bone\n"
		if LinkType == 2 do format "gmrootbone\n"
	)
)


/* Vertex struct for final surface verts */
struct gmVertex
(
	position=[0,0,0],
	uvw=[0,0,0],
	normal=[0,0,0],
	tangent=[0,0,0],
	bitangent=[0,0,0],
	boneindex=#(),
	boneweight=#()
)
/* All vertex normals per vertex. Can be more than 1 group since multiple faces can use a single vert */
struct gmVertexNormals
(
	sgroup=#(),
	faces=#(),
	normal=#(),
	tangent=#(),
	bitangent=#()
)
/* Finalized surface struct */
struct gmSurface
(
	NumVerts,
	Vertex=#(),		-- holds gmVertex structs
	NumTris,
	Tris=#(),			-- list of point3's
	Material,
	
	fn PrintFullSurface= --- fuck you crazy man? dont do this!
	(
		format "Numverts: %\n" NumVerts
		for i = 1 to NumVerts do
		(
			--format "Vert: % \tPos: %\n\t\t    Uvw: %\n" i Vertex[i].position Vertex[i].uvw
			format "Vert: %\tPos: %,%,%\n" i (Vertex[i].position[1] as integer) (Vertex[i].position[2] as integer) (Vertex[i].position[3] as integer)
			format "\t%\t%\n" Vertex[i].boneindex Vertex[i].boneweight
			--format "\t\tNormals: %\t\t%\t\t%\n" Vertex[i].normal Vertex[i].tangent Vertex[i].bitangent
		)
		format "NumTris: %\n" NumTris
		for i = 1 to NumTris do
		(
			format "Tri %\t%\n" i Tris[i]
		)
	)
)

/* gm material */
struct gmMaterial
(
	isDummy=true,
	isMultiMaterial=false,
	maxMaterial,

	fn DebugOut=
	(
		format "\t\tisDummy=%\n" isDummy
		format "\t\tisMultiMaterial=%\n" isMultiMaterial
		format "\t\tMaxMaterial:\n"
		format "\t\t\t%\n" maxMaterial
	)
)

/* list of ordered bones */
struct gmHierarchy
(
	isHierarchy=false,
	count=0,
	bones=#(),
	bonename=#(),
	parentindex=#(),

	fn DebugOut=
	(
	)
)